import json
from base64 import b64decode
from datetime import datetime, timezone

from ..exceptions import AnymailConfigurationError
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from ..utils import get_anymail_setting
from .base import AnymailBaseWebhookView


class SparkPostBaseWebhookView(AnymailBaseWebhookView):
    """Base view class for SparkPost webhooks"""

    esp_name = "SparkPost"

    def parse_events(self, request):
        raw_events = json.loads(request.body.decode("utf-8"))
        unwrapped_events = [self.unwrap_event(raw_event) for raw_event in raw_events]
        return [
            self.esp_to_anymail_event(event_class, event, raw_event)
            for (event_class, event, raw_event) in unwrapped_events
            if event is not None  # filter out empty "ping" events
        ]

    def unwrap_event(self, raw_event):
        """Unwraps SparkPost event structure, and returns event_class, event, raw_event

        raw_event is of form {'msys': {event_class: {...event...}}}

        Can return None, None, raw_event for SparkPost "ping" raw_event={'msys': {}}
        """
        event_classes = raw_event["msys"].keys()
        try:
            (event_class,) = event_classes
            event = raw_event["msys"][event_class]
        except ValueError:  # too many/not enough event_classes to unpack
            if len(event_classes) == 0:
                # Empty event (SparkPost sometimes sends as a "ping")
                event_class = event = None
            else:
                raise TypeError(
                    "Invalid SparkPost webhook event has multiple event classes: %r"
                    % raw_event
                ) from None
        return event_class, event, raw_event

    def esp_to_anymail_event(self, event_class, event, raw_event):
        raise NotImplementedError()


class SparkPostTrackingWebhookView(SparkPostBaseWebhookView):
    """Handler for SparkPost message, engagement, and generation event webhooks"""

    signal = tracking

    event_types = {
        # Map SparkPost event.type: Anymail normalized type
        "bounce": EventType.BOUNCED,
        "delivery": EventType.DELIVERED,
        "injection": EventType.QUEUED,
        "spam_complaint": EventType.COMPLAINED,
        "out_of_band": EventType.BOUNCED,
        "policy_rejection": EventType.REJECTED,
        "delay": EventType.DEFERRED,
        "click": EventType.CLICKED,
        "open": EventType.OPENED,
        "amp_click": EventType.CLICKED,
        "amp_open": EventType.OPENED,
        "generation_failure": EventType.FAILED,
        "generation_rejection": EventType.REJECTED,
        "list_unsubscribe": EventType.UNSUBSCRIBED,
        "link_unsubscribe": EventType.UNSUBSCRIBED,
    }

    # Additional event_types mapping when Anymail setting
    # SPARKPOST_TRACK_INITIAL_OPEN_AS_OPENED is enabled.
    initial_open_event_types = {
        "initial_open": EventType.OPENED,
        "amp_initial_open": EventType.OPENED,
    }

    reject_reasons = {
        # Map SparkPost event.bounce_class: Anymail normalized reject reason.
        # Can also supply (RejectReason, EventType) for bounce_class that affects our
        # event_type. https://support.sparkpost.com/customer/portal/articles/1929896
        "1": RejectReason.OTHER,  # Undetermined (response text could not be identified)
        "10": RejectReason.INVALID,  # Invalid Recipient
        "20": RejectReason.BOUNCED,  # Soft Bounce
        "21": RejectReason.BOUNCED,  # DNS Failure
        "22": RejectReason.BOUNCED,  # Mailbox Full
        "23": RejectReason.BOUNCED,  # Too Large
        "24": RejectReason.TIMED_OUT,  # Timeout
        "25": RejectReason.BLOCKED,  # Admin Failure (configured policies)
        "30": RejectReason.BOUNCED,  # Generic Bounce: No RCPT
        "40": RejectReason.BOUNCED,  # Generic Bounce: unspecified reasons
        "50": RejectReason.BLOCKED,  # Mail Block (by the receiver)
        "51": RejectReason.SPAM,  # Spam Block (by the receiver)
        "52": RejectReason.SPAM,  # Spam Content (by the receiver)
        "53": RejectReason.OTHER,  # Prohibited Attachment (by the receiver)
        "54": RejectReason.BLOCKED,  # Relaying Denied (by the receiver)
        "60": (RejectReason.OTHER, EventType.AUTORESPONDED),  # Auto-Reply/vacation
        "70": RejectReason.BOUNCED,  # Transient Failure
        "80": (RejectReason.OTHER, EventType.SUBSCRIBED),  # Subscribe
        "90": (RejectReason.UNSUBSCRIBED, EventType.UNSUBSCRIBED),  # Unsubscribe
        "100": (RejectReason.OTHER, EventType.AUTORESPONDED),  # Challenge-Response
    }

    def __init__(self, **kwargs):
        # Set Anymail setting SPARKPOST_TRACK_INITIAL_OPEN_AS_OPENED True
        # to report *both* "open" and "initial_open" as Anymail "opened" events.
        # (Otherwise only "open" maps to "opened", matching the behavior of most
        # other ESPs.) Handling "initial_open" is opt-in, to help avoid duplicate
        # "opened" events on the same first open.
        track_initial_open_as_opened = get_anymail_setting(
            "track_initial_open_as_opened",
            default=False,
            esp_name=self.esp_name,
            kwargs=kwargs,
        )
        if track_initial_open_as_opened:
            self.event_types = {**self.event_types, **self.initial_open_event_types}
        super().__init__(**kwargs)

    def esp_to_anymail_event(self, event_class, event, raw_event):
        if event_class == "relay_message":
            # This is an inbound event
            raise AnymailConfigurationError(
                "You seem to have set SparkPost's *inbound* relay webhook URL "
                "to Anymail's SparkPost *tracking* webhook URL."
            )

        event_type = self.event_types.get(event["type"], EventType.UNKNOWN)
        try:
            timestamp = datetime.fromtimestamp(int(event["timestamp"]), tz=timezone.utc)
        except (KeyError, TypeError, ValueError):
            timestamp = None

        try:
            tag = event["campaign_id"]
            # not "rcpt_tags" -- those don't come from sending a message
            tags = [tag] if tag else None
        except KeyError:
            tags = []

        try:
            reject_reason = self.reject_reasons.get(
                event["bounce_class"], RejectReason.OTHER
            )
            try:
                # unpack (RejectReason, EventType)
                # for reasons that change our event type
                reject_reason, event_type = reject_reason
            except ValueError:
                pass
        except KeyError:
            reject_reason = None  # no bounce_class

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            # use transmission_id, not message_id -- see SparkPost backend
            message_id=event.get("transmission_id", None),
            event_id=event.get("event_id", None),
            # raw_rcpt_to preserves email case (vs. rcpt_to)
            recipient=event.get("raw_rcpt_to", None),
            reject_reason=reject_reason,
            mta_response=event.get("raw_reason", None),
            # description=???,
            tags=tags,
            # metadata includes message + recipient metadata
            metadata=event.get("rcpt_meta", None) or {},
            click_url=event.get("target_link_url", None),
            user_agent=event.get("user_agent", None),
            esp_event=raw_event,
        )


class SparkPostInboundWebhookView(SparkPostBaseWebhookView):
    """Handler for SparkPost inbound relay webhook"""

    signal = inbound

    def esp_to_anymail_event(self, event_class, event, raw_event):
        if event_class != "relay_message":
            # This is not an inbound event
            raise AnymailConfigurationError(
                "You seem to have set SparkPost's *tracking* webhook URL "
                "to Anymail's SparkPost *inbound* relay webhook URL."
            )

        if event["protocol"] != "smtp":
            raise AnymailConfigurationError(
                "You cannot use Anymail's webhooks for SparkPost '{protocol}' relay"
                " events. Anymail only handles the 'smtp' protocol".format(
                    protocol=event["protocol"]
                )
            )

        raw_mime = event["content"]["email_rfc822"]
        if event["content"]["email_rfc822_is_base64"]:
            raw_mime = b64decode(raw_mime).decode("utf-8")
        message = AnymailInboundMessage.parse_raw_mime(raw_mime)

        message.envelope_sender = event.get("msg_from", None)
        message.envelope_recipient = event.get("rcpt_to", None)

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            # SparkPost does not provide a relay event timestamp
            timestamp=None,
            # SparkPost does not provide an idempotent id for relay events
            event_id=None,
            esp_event=raw_event,
            message=message,
        )



# Hatch custom build hook that generates dynamic readme.

import re
from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface


def freeze_readme_versions(text: str, version: str) -> str:
    """
    Rewrite links in readme text to refer to specific version.
    (This assumes version X.Y will be tagged "vX.Y" in git.)
    """
    release_tag = f"v{version}"
    return re.sub(
        # (?<=...) is "positive lookbehind": must be there, but won't get replaced
        # GitHub Actions build status: branch=main --> branch=vX.Y.Z:
        r"(?<=branch[=:])main"
        # ReadTheDocs links: /stable --> /vX.Y.Z:
        r"|(?<=/)stable"
        # ReadTheDocs badge: version=stable --> version=vX.Y.Z:
        r"|(?<=version=)stable",
        release_tag,
        text,
    )


class CustomMetadataHook(MetadataHookInterface):
    def update(self, metadata):
        """
        Update the project table's metadata.
        """
        readme_path = Path(self.root) / self.config["readme"]
        content_type = self.config.get("content-type", "text/x-rst")
        version = metadata["version"]

        readme_text = readme_path.read_text()
        readme_text = freeze_readme_versions(readme_text, version)

        metadata["readme"] = {
            "content-type": content_type,
            "text": readme_text,
        }



import mimetypes
from email.charset import QP, Charset
from email.header import decode_header, make_header
from email.headerregistry import Address

from ..message import AnymailRecipientStatus
from ..utils import (
    BASIC_NUMERIC_TYPES,
    CaseInsensitiveCasePreservingDict,
    get_anymail_setting,
)
from .base_requests import AnymailRequestsBackend, RequestsPayload

# Used to force RFC-2047 encoded word
# in address formatting workaround
QP_CHARSET = Charset("utf-8")
QP_CHARSET.header_encoding = QP


class EmailBackend(AnymailRequestsBackend):
    """
    Resend (resend.com) API Email Backend
    """

    esp_name = "Resend"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.api_key = get_anymail_setting(
            "api_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default="https://api.resend.com/",
        )
        if not api_url.endswith("/"):
            api_url += "/"

        # Undocumented setting to control workarounds for Resend display-name issues
        # (see below). If/when Resend improves their API, you can disable Anymail's
        # workarounds by adding `"RESEND_WORKAROUND_DISPLAY_NAME_BUGS": False`
        # to your `ANYMAIL` settings.
        self.workaround_display_name_bugs = get_anymail_setting(
            "workaround_display_name_bugs",
            esp_name=esp_name,
            kwargs=kwargs,
            default=True,
        )

        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return ResendPayload(message, defaults, self)

    def parse_recipient_status(self, response, payload, message):
        # Resend provides single message id, no other information.
        # Assume "queued".
        parsed_response = self.deserialize_json_response(response, payload, message)
        message_id = parsed_response["id"]
        recipient_status = CaseInsensitiveCasePreservingDict(
            {
                recip.addr_spec: AnymailRecipientStatus(
                    message_id=message_id, status="queued"
                )
                for recip in payload.recipients
            }
        )
        return dict(recipient_status)


class ResendPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        self.recipients = []  # for parse_recipient_status
        headers = kwargs.pop("headers", {})
        headers["Authorization"] = "Bearer %s" % backend.api_key
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        super().__init__(message, defaults, backend, headers=headers, *args, **kwargs)

    def get_api_endpoint(self):
        return "emails"

    def serialize_data(self):
        return self.serialize_json(self.data)

    #
    # Payload construction
    #

    def init_payload(self):
        self.data = {}  # becomes json

    def _resend_email_address(self, address):
        """
        Return EmailAddress address formatted for use with Resend.

        Works around a Resend bug that rejects properly formatted RFC 5322
        addresses that have the display-name enclosed in double quotes (e.g.,
        any display-name containing a comma), by substituting an RFC 2047
        encoded word.

        This works for all Resend address fields _except_ `from` (see below).
        """
        formatted = address.address
        if self.backend.workaround_display_name_bugs:
            if formatted.startswith('"'):
                # Workaround: force RFC-2047 encoded word
                formatted = str(
                    Address(
                        display_name=QP_CHARSET.header_encode(address.display_name),
                        addr_spec=address.addr_spec,
                    )
                )
        return formatted

    def set_from_email(self, email):
        # Can't use the address header workaround above for the `from` field:
        #   self.data["from"] = self._resend_email_address(email)
        # When `from` uses RFC-2047 encoding, Resend returns a "security_error"
        # status 451, "The email payload contain invalid characters".
        formatted = email.address
        if self.backend.workaround_display_name_bugs:
            if formatted.startswith("=?"):
                # Workaround: use an *unencoded* (Unicode str) display-name.
                # This allows use of non-ASCII characters (which Resend rejects when
                # encoded with RFC 2047). Some punctuation will still result in unusual
                # behavior or cause an "invalid `from` field" 422 error, but there's
                # nothing we can do about that.
                formatted = str(
                    # email.headerregistry.Address str format uses unencoded Unicode
                    Address(
                        # Convert RFC 2047 display name back to Unicode str
                        display_name=str(
                            make_header(decode_header(email.display_name))
                        ),
                        addr_spec=email.addr_spec,
                    )
                )
        self.data["from"] = formatted

    def set_recipients(self, recipient_type, emails):
        assert recipient_type in ["to", "cc", "bcc"]
        if emails:
            field = recipient_type
            self.data[field] = [self._resend_email_address(email) for email in emails]
            self.recipients += emails

    def set_subject(self, subject):
        self.data["subject"] = subject

    def set_reply_to(self, emails):
        if emails:
            self.data["reply_to"] = [
                self._resend_email_address(email) for email in emails
            ]

    def set_extra_headers(self, headers):
        # Resend requires header values to be strings (not integers) as of 2023-10-20.
        # Stringify ints and floats; anything else is the caller's responsibility.
        self.data.setdefault("headers", {}).update(
            {
                k: str(v) if isinstance(v, BASIC_NUMERIC_TYPES) else v
                for k, v in headers.items()
            }
        )

    def set_text_body(self, body):
        self.data["text"] = body

    def set_html_body(self, body):
        if "html" in self.data:
            # second html body could show up through multiple alternatives,
            # or html body + alternative
            self.unsupported_feature("multiple html parts")
        self.data["html"] = body

    @staticmethod
    def make_attachment(attachment):
        """Returns Resend attachment dict for attachment"""
        filename = attachment.name or ""
        if not filename:
            # Provide default name with reasonable extension.
            # (Resend guesses content type from the filename extension;
            # there doesn't seem to be any other way to specify it.)
            ext = mimetypes.guess_extension(attachment.content_type)
            if ext is not None:
                filename = f"attachment{ext}"
        att = {"content": attachment.b64content, "filename": filename}
        # attachment.inline / attachment.cid not supported
        return att

    def set_attachments(self, attachments):
        if attachments:
            if any(att.content_id for att in attachments):
                self.unsupported_feature("inline content-id")
            self.data["attachments"] = [
                self.make_attachment(attachment) for attachment in attachments
            ]

    def set_metadata(self, metadata):
        # Send metadata as json in a custom X-Metadata header.
        # (Resend's own "tags" are severely limited in character set)
        self.data.setdefault("headers", {})["X-Metadata"] = self.serialize_json(
            metadata
        )

    # Resend doesn't support delayed sending
    # def set_send_at(self, send_at):

    def set_tags(self, tags):
        # Send tags using a custom X-Tags header.
        # (Resend's own "tags" are severely limited in character set)
        self.data.setdefault("headers", {})["X-Tags"] = self.serialize_json(tags)

    # Resend doesn't support changing click/open tracking per message
    # def set_track_clicks(self, track_clicks):
    # def set_track_opens(self, track_opens):

    # Resend doesn't support server-rendered templates.
    # (Their template feature is rendered client-side,
    # using React in node.js.)
    # def set_template_id(self, template_id):
    # def set_merge_data(self, merge_data):
    # def set_merge_global_data(self, merge_global_data):
    # def set_merge_metadata(self, merge_metadata):

    def set_esp_extra(self, extra):
        self.data.update(extra)



import hashlib
import hmac
import json
from base64 import b64encode
from datetime import datetime, timezone

from django.utils.crypto import constant_time_compare

from ..exceptions import AnymailWebhookValidationFailure
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    inbound,
    tracking,
)
from ..utils import get_anymail_setting, get_request_uri, getfirst
from .base import AnymailBaseWebhookView, AnymailCoreWebhookView


class MandrillSignatureMixin(AnymailCoreWebhookView):
    """Validates Mandrill webhook signature"""

    # These can be set from kwargs in View.as_view, or pulled from settings in init:
    webhook_key = None  # required
    webhook_url = None  # optional; defaults to actual url used

    def __init__(self, **kwargs):
        esp_name = self.esp_name
        # webhook_key is required for POST, but not for HEAD when Mandrill validates
        # webhook url. Defer "missing setting" error until we actually try to use it in
        # the POST...
        webhook_key = get_anymail_setting(
            "webhook_key",
            esp_name=esp_name,
            default=None,
            kwargs=kwargs,
            allow_bare=True,
        )
        if webhook_key is not None:
            # hmac.new requires bytes key
            self.webhook_key = webhook_key.encode("ascii")
        self.webhook_url = get_anymail_setting(
            "webhook_url",
            esp_name=esp_name,
            default=None,
            kwargs=kwargs,
            allow_bare=True,
        )
        super().__init__(**kwargs)

    def validate_request(self, request):
        if self.webhook_key is None:
            # issue deferred "missing setting" error
            # (re-call get-setting without a default)
            get_anymail_setting("webhook_key", esp_name=self.esp_name, allow_bare=True)

        try:
            signature = request.META["HTTP_X_MANDRILL_SIGNATURE"]
        except KeyError:
            raise AnymailWebhookValidationFailure(
                "X-Mandrill-Signature header missing from webhook POST"
            ) from None

        # Mandrill signs the exact URL (including basic auth, if used)
        # plus the sorted POST params:
        url = self.webhook_url or get_request_uri(request)
        params = request.POST.dict()
        signed_data = url
        for key in sorted(params.keys()):
            signed_data += key + params[key]

        expected_signature = b64encode(
            hmac.new(
                key=self.webhook_key,
                msg=signed_data.encode("utf-8"),
                digestmod=hashlib.sha1,
            ).digest()
        )
        if not constant_time_compare(signature, expected_signature):
            raise AnymailWebhookValidationFailure(
                "Mandrill webhook called with incorrect signature (for url %r)" % url
            )


class MandrillCombinedWebhookView(MandrillSignatureMixin, AnymailBaseWebhookView):
    """Unified view class for Mandrill tracking and inbound webhooks"""

    esp_name = "Mandrill"

    warn_if_no_basic_auth = False  # because we validate against signature
    signal = None  # set in esp_to_anymail_event

    def parse_events(self, request):
        esp_events = json.loads(request.POST["mandrill_events"])
        return [self.esp_to_anymail_event(esp_event) for esp_event in esp_events]

    def esp_to_anymail_event(self, esp_event):
        """Route events to the inbound or tracking handler"""
        esp_type = getfirst(esp_event, ["event", "type"], "unknown")

        if esp_type == "inbound":
            assert self.signal is not tracking  # batch must not mix event types
            self.signal = inbound
            return self.mandrill_inbound_to_anymail_event(esp_event)
        else:
            assert self.signal is not inbound  # batch must not mix event types
            self.signal = tracking
            return self.mandrill_tracking_to_anymail_event(esp_event)

    #
    #  Tracking events
    #

    event_types = {
        # Message events:
        "send": EventType.SENT,
        "deferral": EventType.DEFERRED,
        "hard_bounce": EventType.BOUNCED,
        "soft_bounce": EventType.BOUNCED,
        "open": EventType.OPENED,
        "click": EventType.CLICKED,
        "spam": EventType.COMPLAINED,
        "unsub": EventType.UNSUBSCRIBED,
        "reject": EventType.REJECTED,
        # Sync events (we don't really normalize these well):
        "whitelist": EventType.UNKNOWN,
        "blacklist": EventType.UNKNOWN,
        # Inbound events:
        "inbound": EventType.INBOUND,
    }

    def mandrill_tracking_to_anymail_event(self, esp_event):
        esp_type = getfirst(esp_event, ["event", "type"], None)
        event_type = self.event_types.get(esp_type, EventType.UNKNOWN)

        try:
            timestamp = datetime.fromtimestamp(esp_event["ts"], tz=timezone.utc)
        except (KeyError, ValueError):
            timestamp = None

        try:
            recipient = esp_event["msg"]["email"]
        except KeyError:
            try:
                recipient = esp_event["reject"]["email"]  # sync events
            except KeyError:
                recipient = None

        try:
            mta_response = esp_event["msg"]["diag"]
        except KeyError:
            mta_response = None

        try:
            description = getfirst(esp_event["reject"], ["detail", "reason"])
        except KeyError:
            description = None

        try:
            metadata = esp_event["msg"]["metadata"]
        except KeyError:
            metadata = {}

        try:
            tags = esp_event["msg"]["tags"]
        except KeyError:
            tags = []

        return AnymailTrackingEvent(
            click_url=esp_event.get("url", None),
            description=description,
            esp_event=esp_event,
            event_type=event_type,
            message_id=esp_event.get("_id", None),
            metadata=metadata,
            mta_response=mta_response,
            recipient=recipient,
            # reject_reason should probably map esp_event['msg']['bounce_description'],
            # but Mandrill docs are insufficient to determine how
            reject_reason=None,
            tags=tags,
            timestamp=timestamp,
            user_agent=esp_event.get("user_agent", None),
        )

    #
    # Inbound events
    #

    def mandrill_inbound_to_anymail_event(self, esp_event):
        # It's easier (and more accurate) to just work
        # from the original raw mime message
        message = AnymailInboundMessage.parse_raw_mime(esp_event["msg"]["raw_msg"])

        # (Mandrill's "sender" field only applies to outbound messages)
        message.envelope_sender = None
        message.envelope_recipient = esp_event["msg"].get("email", None)

        # no simple boolean spam; would need to parse the spam_report
        message.spam_detected = None
        message.spam_score = esp_event["msg"].get("spam_report", {}).get("score", None)

        try:
            timestamp = datetime.fromtimestamp(esp_event["ts"], tz=timezone.utc)
        except (KeyError, ValueError):
            timestamp = None

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=timestamp,
            # Mandrill doesn't provide an idempotent inbound message event id
            event_id=None,
            esp_event=esp_event,
            message=message,
        )


# Backwards-compatibility:
# earlier Anymail versions had only MandrillTrackingWebhookView:
MandrillTrackingWebhookView = MandrillCombinedWebhookView



# Don't import this file directly (unless you are a build system).
# Instead, load version info from the package root.

#: major.minor or major.minor.patch (optionally with .devN suffix)
__version__ = "10.2"

VERSION = __version__.split(",")



from ._version import VERSION, __version__

__all__ = [
    "VERSION",
    "__version__",
]

try:
    import django
except ImportError:
    # (don't require django just to get package version)
    pass
else:
    if django.VERSION < (3, 2, 0):
        # (No longer required -- and causes deprecation warning -- in Django 3.2+)
        default_app_config = "anymail.apps.AnymailBaseConfig"
        __all__.append("default_app_config")



from django.dispatch import Signal

#: Outbound message, before sending
#: provides args: message, esp_name
pre_send = Signal()

#: Outbound message, after sending
#: provides args: message, status, esp_name
post_send = Signal()

#: Delivery and tracking events for sent messages
#: provides args: event, esp_name
tracking = Signal()

#: Event for receiving inbound messages
#: provides args: event, esp_name
inbound = Signal()


class AnymailEvent:
    """Base class for normalized Anymail webhook events"""

    def __init__(
        self, event_type, timestamp=None, event_id=None, esp_event=None, **kwargs
    ):
        #: normalized to an EventType str
        self.event_type = event_type
        #: normalized to an aware datetime
        self.timestamp = timestamp
        #: opaque str
        self.event_id = event_id
        #: raw event fields (e.g., parsed JSON dict or POST data QueryDict)
        self.esp_event = esp_event


class AnymailTrackingEvent(AnymailEvent):
    """Normalized delivery and tracking event for sent messages"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.click_url = kwargs.pop("click_url", None)  #: str
        #: str, usually human-readable, not normalized
        self.description = kwargs.pop("description", None)
        self.message_id = kwargs.pop("message_id", None)  #: str, format may vary
        self.metadata = kwargs.pop("metadata", {})  #: dict
        #: str, may include SMTP codes, not normalized
        self.mta_response = kwargs.pop("mta_response", None)
        #: str email address (just the email portion; no name)
        self.recipient = kwargs.pop("recipient", None)
        #: normalized to a RejectReason str
        self.reject_reason = kwargs.pop("reject_reason", None)
        self.tags = kwargs.pop("tags", [])  #: list of str
        self.user_agent = kwargs.pop("user_agent", None)  #: str


class AnymailInboundEvent(AnymailEvent):
    """Normalized inbound message event"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        #: anymail.inbound.AnymailInboundMessage
        self.message = kwargs.pop("message", None)


class EventType:
    """Constants for normalized Anymail event types"""

    # Delivery (and non-delivery) event types
    # (these match message.ANYMAIL_STATUSES where appropriate)

    #: the ESP has accepted the message and will try to send it (possibly later)
    QUEUED = "queued"

    #: the ESP has sent the message (though it may or may not get delivered)
    SENT = "sent"

    #: the ESP refused to send the message
    #: (e.g., suppression list, policy, invalid email)
    REJECTED = "rejected"

    #: the ESP was unable to send the message (e.g., template rendering error)
    FAILED = "failed"

    #: rejected or blocked by receiving MTA
    BOUNCED = "bounced"

    #: delayed by receiving MTA; should be followed by a later BOUNCED or DELIVERED
    DEFERRED = "deferred"

    #: accepted by receiving MTA
    DELIVERED = "delivered"

    #: a bot replied
    AUTORESPONDED = "autoresponded"

    # Tracking event types

    #: open tracking
    OPENED = "opened"

    #: click tracking
    CLICKED = "clicked"

    #: recipient reported as spam (e.g., through feedback loop)
    COMPLAINED = "complained"

    #: recipient attempted to unsubscribe
    UNSUBSCRIBED = "unsubscribed"

    #: signed up for mailing list through ESP-hosted form
    SUBSCRIBED = "subscribed"

    # Inbound event types

    #: received message
    INBOUND = "inbound"

    #: (ESP notification of) error receiving message
    INBOUND_FAILED = "inbound_failed"

    # Other event types

    #: all other ESP events
    UNKNOWN = "unknown"


class RejectReason:
    """Constants for normalized Anymail reject/drop reasons"""

    #: bad address format
    INVALID = "invalid"

    #: (previous) bounce from recipient
    BOUNCED = "bounced"

    #: (previous) repeated failed delivery attempts
    TIMED_OUT = "timed_out"

    #: ESP policy suppression
    BLOCKED = "blocked"

    #: (previous) spam complaint from recipient
    SPAM = "spam"

    #: (previous) unsubscribe request from recipient
    UNSUBSCRIBED = "unsubscribed"

    #: all other ESP reject reasons
    OTHER = "other"



import hashlib
import hmac
import json

from django.utils.crypto import constant_time_compare
from django.utils.dateparse import parse_datetime

from ..exceptions import AnymailConfigurationError, AnymailWebhookValidationFailure
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from ..utils import get_anymail_setting
from .base import AnymailBaseWebhookView


class MailerSendBaseWebhookView(AnymailBaseWebhookView):
    """Base view class for MailerSend webhooks"""

    esp_name = "MailerSend"
    warn_if_no_basic_auth = False  # because we validate against signature

    def __init__(self, _secret_name, **kwargs):
        signing_secret = get_anymail_setting(
            _secret_name,
            esp_name=self.esp_name,
            kwargs=kwargs,
        )
        # hmac.new requires bytes key:
        self.signing_secret = signing_secret.encode("ascii")
        self._secret_setting_name = f"{self.esp_name}_{_secret_name}".upper()
        super().__init__(**kwargs)

    def validate_request(self, request):
        super().validate_request(request)  # first check basic auth if enabled
        try:
            signature = request.headers["Signature"]
        except KeyError:
            raise AnymailWebhookValidationFailure(
                "MailerSend webhook called without signature"
            ) from None

        expected_signature = hmac.new(
            key=self.signing_secret,
            msg=request.body,
            digestmod=hashlib.sha256,
        ).hexdigest()
        if not constant_time_compare(signature, expected_signature):
            raise AnymailWebhookValidationFailure(
                f"MailerSend webhook called with incorrect signature"
                f" (check Anymail {self._secret_setting_name} setting)"
            )


class MailerSendTrackingWebhookView(MailerSendBaseWebhookView):
    """Handler for MailerSend delivery and engagement tracking webhooks"""

    signal = tracking

    # (Declaring class attr allows override by kwargs in View.as_view.)
    signing_secret = None

    def __init__(self, **kwargs):
        super().__init__(_secret_name="signing_secret", **kwargs)

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))
        event_type = esp_event.get("type")
        if event_type == "inbound.message":
            raise AnymailConfigurationError(
                "You seem to have set MailerSend's *inbound* route endpoint"
                " to Anymail's MailerSend *activity tracking* webhook URL. "
            )
        return [self.esp_to_anymail_event(esp_event)]

    event_types = {
        # Map MailerSend activity.type: Anymail normalized type
        "sent": EventType.SENT,
        "delivered": EventType.DELIVERED,
        "soft_bounced": EventType.BOUNCED,
        "hard_bounced": EventType.BOUNCED,
        "opened": EventType.OPENED,
        "clicked": EventType.CLICKED,
        "unsubscribed": EventType.UNSUBSCRIBED,
        "spam_complaint": EventType.COMPLAINED,
    }

    morph_reject_reasons = {
        # Map MailerSend morph.object (type): Anymail normalized RejectReason
        "recipient_bounce": RejectReason.BOUNCED,
        "spam_complaint": RejectReason.SPAM,
        "recipient_unsubscribe": RejectReason.UNSUBSCRIBED,
        # any others?
    }

    def esp_to_anymail_event(self, esp_event):
        activity_data = esp_event.get("data", {})
        email_data = activity_data.get("email", {})
        message_data = email_data.get("message", {})
        recipient_data = email_data.get("recipient", {})

        event_type = self.event_types.get(activity_data["type"], EventType.UNKNOWN)
        event_id = activity_data.get("id")
        recipient = recipient_data.get("email")
        message_id = message_data.get("id")
        tags = email_data.get("tags", [])

        try:
            timestamp = parse_datetime(activity_data["created_at"])
        except KeyError:
            timestamp = None

        # Additional, event-specific info is included in a "morph" record.
        try:
            morph_data = activity_data["morph"]
            morph_object = morph_data["object"]  # the object type of morph_data
        except (KeyError, TypeError):
            reject_reason = None
            description = None
            click_url = None
        else:
            # It seems like email_data["status"] should map to a reject_reason, but in
            # reality status is most often just (the undocumented) "rejected" and the
            # morph_object has more accurate info.
            reject_reason = self.morph_reject_reasons.get(morph_object)
            description = morph_data.get("readable_reason") or morph_data.get("reason")
            click_url = morph_data.get("url")  # object="click"
            # user_ip = morph_data.get("ip")  # object="click" or "open"

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            message_id=message_id,
            event_id=event_id,
            recipient=recipient,
            reject_reason=reject_reason,
            description=description,
            tags=tags,
            click_url=click_url,
            esp_event=esp_event,
        )


class MailerSendInboundWebhookView(MailerSendBaseWebhookView):
    """Handler for MailerSend inbound webhook"""

    signal = inbound

    # (Declaring class attr allows override by kwargs in View.as_view.)
    inbound_secret = None

    def __init__(self, **kwargs):
        super().__init__(_secret_name="inbound_secret", **kwargs)

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))
        event_type = esp_event.get("type")
        if event_type != "inbound.message":
            raise AnymailConfigurationError(
                f"You seem to have set MailerSend's *{event_type}* webhook "
                "to Anymail's MailerSend *inbound* webhook URL. "
            )
        return [self.esp_to_anymail_event(esp_event)]

    def esp_to_anymail_event(self, esp_event):
        message_data = esp_event.get("data")
        event_id = message_data.get("id")

        try:
            timestamp = parse_datetime(message_data["created_at"])
        except (KeyError, TypeError):
            timestamp = None

        message = AnymailInboundMessage.parse_raw_mime(message_data.get("raw"))

        try:
            message.envelope_sender = message_data["sender"]["email"]
            # (also available as X-Envelope-From header)
        except KeyError:
            pass

        try:
            # There can be multiple rcptTo if the same message is sent
            # to multiple inbound recipients. Just use the first.
            envelope_recipients = [
                recipient["email"] for recipient in message_data["recipients"]["rcptTo"]
            ]
            message.envelope_recipient = envelope_recipients[0]
        except (KeyError, IndexError):
            pass

        # MailerSend doesn't seem to provide any spam annotations.
        # SPF seems to be verified, but format is undocumented:
        #     "spf_check": {"code": "+", "value": None}
        # DKIM doesn't appear to be verified yet:
        #     "dkim_check": False,

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=timestamp,
            event_id=event_id,
            esp_event=esp_event,
            message=message,
        )



from urllib.parse import urljoin

import requests

from anymail.utils import get_anymail_setting

from .._version import __version__
from ..exceptions import AnymailRequestsAPIError
from .base import AnymailBaseBackend, BasePayload


class AnymailRequestsBackend(AnymailBaseBackend):
    """
    Base Anymail email backend for ESPs that use an HTTP API via requests
    """

    def __init__(self, api_url, **kwargs):
        """Init options from Django settings"""
        self.api_url = api_url
        self.timeout = get_anymail_setting(
            "requests_timeout", kwargs=kwargs, default=30
        )
        super().__init__(**kwargs)
        self.session = None

    def open(self):
        if self.session:
            return False  # already exists

        try:
            self.session = self.create_session()
        except Exception:
            if not self.fail_silently:
                raise

        return True

    def close(self):
        if self.session is None:
            return
        try:
            self.session.close()
        except requests.RequestException:
            if not self.fail_silently:
                raise
        finally:
            self.session = None

    def _send(self, message):
        if self.session:
            return super()._send(message)
        elif self.fail_silently:
            # create_session failed
            return False
        else:
            class_name = self.__class__.__name__
            raise RuntimeError(
                "Session has not been opened in {class_name}._send. "
                "(This is either an implementation error in {class_name}, "
                "or you are incorrectly calling _send directly.)".format(
                    class_name=class_name
                )
            )

    def create_session(self):
        """Create the requests.Session object used by this instance of the backend.
        If subclassed, you can modify the Session returned from super() to give
        it your own configuration.

        This must return an instance of requests.Session."""
        session = requests.Session()

        session.headers["User-Agent"] = "django-anymail/{version}-{esp} {orig}".format(
            esp=self.esp_name.lower(),
            version=__version__,
            orig=session.headers.get("User-Agent", ""),
        )

        if self.debug_api_requests:
            session.hooks["response"].append(self._dump_api_request)

        return session

    def post_to_esp(self, payload, message):
        """Post payload to ESP send API endpoint, and return the raw response.

        payload is the result of build_message_payload
        message is the original EmailMessage
        return should be a requests.Response

        Can raise AnymailRequestsAPIError for HTTP errors in the post
        """
        params = payload.get_request_params(self.api_url)
        params.setdefault("timeout", self.timeout)
        try:
            response = self.session.request(**params)
        except requests.RequestException as err:
            # raise an exception that is both AnymailRequestsAPIError
            # and the original requests exception type
            exc_class = type(
                "AnymailRequestsAPIError", (AnymailRequestsAPIError, type(err)), {}
            )
            raise exc_class(
                "Error posting to %s:" % params.get("url", "<missing url>"),
                email_message=message,
                payload=payload,
            ) from err
        self.raise_for_status(response, payload, message)
        return response

    def raise_for_status(self, response, payload, message):
        """Raise AnymailRequestsAPIError if response is an HTTP error

        Subclasses can override for custom error checking
        (though should defer parsing/deserialization of the body to
        parse_recipient_status)
        """
        if response.status_code < 200 or response.status_code >= 300:
            raise AnymailRequestsAPIError(
                email_message=message, payload=payload, response=response, backend=self
            )

    def deserialize_json_response(self, response, payload, message):
        """Deserialize an ESP API response that's in json.

        Useful for implementing deserialize_response
        """
        try:
            return response.json()
        except ValueError as err:
            raise AnymailRequestsAPIError(
                "Invalid JSON in %s API response" % self.esp_name,
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            ) from err

    @staticmethod
    def _dump_api_request(response, **kwargs):
        """Print the request and response for debugging"""
        # (This is not byte-for-byte, but a readable text representation that assumes
        # UTF-8 encoding if encoded, and that omits the CR in CRLF line endings.
        # If you need the raw bytes, configure HTTPConnection logging as shown
        # in http://docs.python-requests.org/en/v3.0.0/api/#api-changes)
        request = response.request  # a PreparedRequest
        print("\n===== Anymail API request")
        print(
            "{method} {url}\n{headers}".format(
                method=request.method,
                url=request.url,
                headers="".join(
                    "{header}: {value}\n".format(header=header, value=value)
                    for (header, value) in request.headers.items()
                ),
            )
        )
        if request.body is not None:
            body_text = (
                request.body
                if isinstance(request.body, str)
                else request.body.decode("utf-8", errors="replace")
            ).replace("\r\n", "\n")
            print(body_text)
        print("\n----- Response")
        print(
            "HTTP {status} {reason}\n{headers}\n{body}".format(
                status=response.status_code,
                reason=response.reason,
                headers="".join(
                    "{header}: {value}\n".format(header=header, value=value)
                    for (header, value) in response.headers.items()
                ),
                body=response.text,  # Let Requests decode body content for us
            )
        )


class RequestsPayload(BasePayload):
    """Abstract Payload for AnymailRequestsBackend"""

    def __init__(
        self,
        message,
        defaults,
        backend,
        method="POST",
        params=None,
        data=None,
        headers=None,
        files=None,
        auth=None,
    ):
        self.method = method
        self.params = params
        self.data = data
        self.headers = headers
        self.files = files
        self.auth = auth
        super().__init__(message, defaults, backend)

    def get_request_params(self, api_url):
        """Returns a dict of requests.request params that will send payload to the ESP.

        :param api_url: the base api_url for the backend
        :return: dict
        """
        api_endpoint = self.get_api_endpoint()
        if api_endpoint is not None:
            url = urljoin(api_url, api_endpoint)
        else:
            url = api_url

        return dict(
            method=self.method,
            url=url,
            params=self.params,
            data=self.serialize_data(),
            headers=self.headers,
            files=self.files,
            auth=self.auth,
            # json= is not here, because we prefer to do our own serialization
            #       to provide extra context in error messages
        )

    def get_api_endpoint(self):
        """
        Returns a str that should be joined to the backend's api_url
        for sending this payload.
        """
        return None

    def serialize_data(self):
        """Performs any necessary serialization on self.data, and returns the result."""
        return self.data



import warnings

from ..exceptions import AnymailDeprecationWarning
from .amazon_ses import EmailBackend as AmazonSESV2EmailBackend


class EmailBackend(AmazonSESV2EmailBackend):
    def __init__(self, **kwargs):
        warnings.warn(
            "Anymail now uses Amazon SES v2 by default. Please change"
            " 'amazon_sesv2' to 'amazon_ses' in your EMAIL_BACKEND setting.",
            AnymailDeprecationWarning,
        )
        super().__init__(**kwargs)



import json
from datetime import datetime, timezone
from email.parser import BytesParser
from email.policy import default as default_policy

from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from .base import AnymailBaseWebhookView


class SendGridTrackingWebhookView(AnymailBaseWebhookView):
    """Handler for SendGrid delivery and engagement tracking webhooks"""

    esp_name = "SendGrid"
    signal = tracking

    def parse_events(self, request):
        esp_events = json.loads(request.body.decode("utf-8"))
        return [self.esp_to_anymail_event(esp_event) for esp_event in esp_events]

    event_types = {
        # Map SendGrid event: Anymail normalized type
        "bounce": EventType.BOUNCED,
        "deferred": EventType.DEFERRED,
        "delivered": EventType.DELIVERED,
        "dropped": EventType.REJECTED,
        "processed": EventType.QUEUED,
        "click": EventType.CLICKED,
        "open": EventType.OPENED,
        "spamreport": EventType.COMPLAINED,
        "unsubscribe": EventType.UNSUBSCRIBED,
        "group_unsubscribe": EventType.UNSUBSCRIBED,
        "group_resubscribe": EventType.SUBSCRIBED,
    }

    reject_reasons = {
        # Map SendGrid reason/type strings (lowercased)
        # to Anymail normalized reject_reason
        "invalid": RejectReason.INVALID,
        "unsubscribed address": RejectReason.UNSUBSCRIBED,
        "bounce": RejectReason.BOUNCED,
        "blocked": RejectReason.BLOCKED,
        "expired": RejectReason.TIMED_OUT,
    }

    def esp_to_anymail_event(self, esp_event):
        event_type = self.event_types.get(esp_event["event"], EventType.UNKNOWN)
        try:
            timestamp = datetime.fromtimestamp(esp_event["timestamp"], tz=timezone.utc)
        except (KeyError, ValueError):
            timestamp = None

        if esp_event["event"] == "dropped":
            # message dropped at ESP before even getting to MTA:
            mta_response = None
            # cause could be in "type" or "reason":
            reason = esp_event.get("type", esp_event.get("reason", ""))
            reject_reason = self.reject_reasons.get(reason.lower(), RejectReason.OTHER)
        else:
            # MTA response is in "response" for delivered; "reason" for bounce
            mta_response = esp_event.get("response", esp_event.get("reason", None))
            reject_reason = None

        # SendGrid merges metadata ('unique_args') with the event.
        # We can (sort of) split metadata back out by filtering known
        # SendGrid event params, though this can miss metadata keys
        # that duplicate SendGrid params, and can accidentally include
        # non-metadata keys if SendGrid modifies their event records.
        metadata_keys = set(esp_event.keys()) - self.sendgrid_event_keys
        if len(metadata_keys) > 0:
            metadata = {key: esp_event[key] for key in metadata_keys}
        else:
            metadata = {}

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            # (smtp-id for backwards compatibility)
            message_id=esp_event.get("anymail_id", esp_event.get("smtp-id")),
            event_id=esp_event.get("sg_event_id", None),
            recipient=esp_event.get("email", None),
            reject_reason=reject_reason,
            mta_response=mta_response,
            tags=esp_event.get("category", []),
            metadata=metadata,
            click_url=esp_event.get("url", None),
            user_agent=esp_event.get("useragent", None),
            esp_event=esp_event,
        )

    # Known keys in SendGrid events (used to recover metadata above)
    sendgrid_event_keys = {
        "anymail_id",
        "asm_group_id",
        "attempt",  # MTA deferred count
        "category",
        "cert_err",
        "email",
        "event",
        "ip",
        "marketing_campaign_id",
        "marketing_campaign_name",
        "newsletter",  # ???
        "nlvx_campaign_id",
        "nlvx_campaign_split_id",
        "nlvx_user_id",
        "pool",
        "post_type",
        "reason",  # MTA bounce/drop reason; SendGrid suppression reason
        "response",  # MTA deferred/delivered message
        "send_at",
        "sg_event_id",
        "sg_message_id",
        "smtp-id",
        "status",  # SMTP status code
        "timestamp",
        "tls",
        "type",  # suppression reject reason ("bounce", "blocked", "expired")
        "url",  # click tracking
        "url_offset",  # click tracking
        "useragent",  # click/open tracking
    }


class SendGridInboundWebhookView(AnymailBaseWebhookView):
    """Handler for SendGrid inbound webhook"""

    esp_name = "SendGrid"
    signal = inbound

    def parse_events(self, request):
        return [self.esp_to_anymail_event(request)]

    def esp_to_anymail_event(self, request):
        # Inbound uses the entire Django request as esp_event, because we need
        # POST and FILES. Note that request.POST is case-sensitive (unlike
        # email.message.Message headers).
        esp_event = request
        # Must access body before any POST fields, or it won't be available if we need
        # it later (see text_charset and html_charset handling below).
        _ensure_body_is_available_later = request.body  # noqa: F841
        if "headers" in request.POST:
            # Default (not "Send Raw") inbound fields
            message = self.message_from_sendgrid_parsed(esp_event)
        elif "email" in request.POST:
            # "Send Raw" full MIME
            message = AnymailInboundMessage.parse_raw_mime(request.POST["email"])
        else:
            raise KeyError(
                "Invalid SendGrid inbound event data"
                " (missing both 'headers' and 'email' fields)"
            )

        try:
            envelope = json.loads(request.POST["envelope"])
        except (KeyError, TypeError, ValueError):
            pass
        else:
            message.envelope_sender = envelope["from"]
            message.envelope_recipient = envelope["to"][0]

        # no simple boolean spam; would need to parse the spam_report
        message.spam_detected = None
        try:
            message.spam_score = float(request.POST["spam_score"])
        except (KeyError, TypeError, ValueError):
            pass

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            # SendGrid doesn't provide an inbound event timestamp:
            timestamp=None,
            # SendGrid doesn't provide an idempotent inbound message event id:
            event_id=None,
            esp_event=esp_event,
            message=message,
        )

    def message_from_sendgrid_parsed(self, request):
        """Construct a Message from SendGrid's "default" (non-raw) fields"""

        try:
            charsets = json.loads(request.POST["charsets"])
        except (KeyError, ValueError):
            charsets = {}

        try:
            attachment_info = json.loads(request.POST["attachment-info"])
        except (KeyError, ValueError):
            attachments = None
        else:
            # Load attachments from posted files
            attachments = []
            for attachment_id in sorted(attachment_info.keys()):
                try:
                    file = request.FILES[attachment_id]
                except KeyError:
                    # Django's multipart/form-data handling drops FILES with certain
                    # filenames (for security) or with empty filenames (Django ticket
                    # 15879). (To avoid this problem, enable SendGrid's "raw, full MIME"
                    # inbound option.)
                    pass
                else:
                    # (This deliberately ignores
                    # attachment_info[attachment_id]["filename"],
                    # which has not passed through Django's filename sanitization.)
                    content_id = attachment_info[attachment_id].get("content-id")
                    attachment = (
                        AnymailInboundMessage.construct_attachment_from_uploaded_file(
                            file, content_id=content_id
                        )
                    )
                    attachments.append(attachment)

        default_charset = request.POST.encoding.lower()  # (probably utf-8)
        text = request.POST.get("text")
        text_charset = charsets.get("text", default_charset).lower()
        html = request.POST.get("html")
        html_charset = charsets.get("html", default_charset).lower()
        if (text and text_charset != default_charset) or (
            html and html_charset != default_charset
        ):
            # Django has parsed text and/or html fields using the wrong charset.
            # We need to re-parse the raw form data and decode each field separately,
            # using the indicated charsets. The email package parses multipart/form-data
            # retaining bytes content. (In theory, we could instead just change
            # request.encoding and access the POST fields again, per Django docs,
            # but that seems to be have bugs around the cached request._files.)
            raw_data = b"".join(
                [
                    b"Content-Type: ",
                    request.META["CONTENT_TYPE"].encode("ascii"),
                    b"\r\n\r\n",
                    request.body,
                ]
            )
            parsed_parts = (
                BytesParser(policy=default_policy).parsebytes(raw_data).get_payload()
            )
            for part in parsed_parts:
                name = part.get_param("name", header="content-disposition")
                if name == "text":
                    text = part.get_payload(decode=True).decode(text_charset)
                elif name == "html":
                    html = part.get_payload(decode=True).decode(html_charset)
                # (subject, from, to, etc. are parsed from raw headers field,
                # so no need to worry about their separate POST field charsets)

        return AnymailInboundMessage.construct(
            # POST["headers"] includes From, To, Cc, Subject, etc.
            raw_headers=request.POST.get("headers", ""),
            text=text,
            html=html,
            attachments=attachments,
        )



import hashlib
import hmac
import json
from datetime import datetime, timezone

from django.utils.crypto import constant_time_compare

from ..exceptions import (
    AnymailConfigurationError,
    AnymailInvalidAddress,
    AnymailWebhookValidationFailure,
)
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from ..utils import (
    UNSET,
    get_anymail_setting,
    merge_dicts_shallow,
    parse_single_address,
    querydict_getfirst,
)
from .base import AnymailBaseWebhookView


class MailgunBaseWebhookView(AnymailBaseWebhookView):
    """Base view class for Mailgun webhooks"""

    esp_name = "Mailgun"
    warn_if_no_basic_auth = False  # because we validate against signature

    # (Declaring class attr allows override by kwargs in View.as_view.)
    webhook_signing_key = None

    # The `api_key` attribute name is still allowed for compatibility
    # with earlier Anymail releases.
    api_key = None  # (Declaring class attr allows override by kwargs in View.as_view.)

    def __init__(self, **kwargs):
        # webhook_signing_key: falls back to api_key if webhook_signing_key not provided
        api_key = get_anymail_setting(
            "api_key",
            esp_name=self.esp_name,
            kwargs=kwargs,
            allow_bare=True,
            default=None,
        )
        webhook_signing_key = get_anymail_setting(
            "webhook_signing_key",
            esp_name=self.esp_name,
            kwargs=kwargs,
            default=UNSET if api_key is None else api_key,
        )
        # hmac.new requires bytes key:
        self.webhook_signing_key = webhook_signing_key.encode("ascii")
        super().__init__(**kwargs)

    def validate_request(self, request):
        super().validate_request(request)  # first check basic auth if enabled
        if request.content_type == "application/json":
            # New-style webhook: json payload with separate signature block
            try:
                event = json.loads(request.body.decode("utf-8"))
                signature_block = event["signature"]
                token = signature_block["token"]
                timestamp = signature_block["timestamp"]
                signature = signature_block["signature"]
            except (KeyError, ValueError, UnicodeDecodeError) as err:
                raise AnymailWebhookValidationFailure(
                    "Mailgun webhook called with invalid payload format"
                ) from err
        else:
            # Legacy webhook: signature fields are interspersed with other POST data
            try:
                # Must use the *last* value of these fields if there are conflicting
                # merged user-variables. (Fortunately, Django QueryDict is specced to
                # return the last value.)
                token = request.POST["token"]
                timestamp = request.POST["timestamp"]
                signature = request.POST["signature"]
            except KeyError as err:
                raise AnymailWebhookValidationFailure(
                    "Mailgun webhook called without required security fields"
                ) from err

        expected_signature = hmac.new(
            key=self.webhook_signing_key,
            msg="{}{}".format(timestamp, token).encode("ascii"),
            digestmod=hashlib.sha256,
        ).hexdigest()
        if not constant_time_compare(signature, expected_signature):
            raise AnymailWebhookValidationFailure(
                "Mailgun webhook called with incorrect signature"
            )


class MailgunTrackingWebhookView(MailgunBaseWebhookView):
    """Handler for Mailgun delivery and engagement tracking webhooks"""

    signal = tracking

    def parse_events(self, request):
        if request.content_type == "application/json":
            esp_event = json.loads(request.body.decode("utf-8"))
            return [self.esp_to_anymail_event(esp_event)]
        else:
            return [self.mailgun_legacy_to_anymail_event(request.POST)]

    event_types = {
        # Map Mailgun event: Anymail normalized type
        "accepted": EventType.QUEUED,  # not delivered to webhooks (8/2018)
        "rejected": EventType.REJECTED,
        "delivered": EventType.DELIVERED,
        "failed": EventType.BOUNCED,
        "opened": EventType.OPENED,
        "clicked": EventType.CLICKED,
        "unsubscribed": EventType.UNSUBSCRIBED,
        "complained": EventType.COMPLAINED,
    }

    reject_reasons = {
        # Map Mailgun event_data.reason: Anymail normalized RejectReason (these appear
        # in webhook doc examples, but aren't actually documented anywhere)
        "bounce": RejectReason.BOUNCED,
        "suppress-bounce": RejectReason.BOUNCED,
        # ??? "generic" appears to be used for any temporary failure?
        "generic": RejectReason.OTHER,
    }

    severities = {
        # Remap some event types based on "severity" payload field
        (EventType.BOUNCED, "temporary"): EventType.DEFERRED
    }

    def esp_to_anymail_event(self, esp_event):
        event_data = esp_event.get("event-data", {})

        event_type = self.event_types.get(event_data["event"], EventType.UNKNOWN)

        event_type = self.severities.get(
            (EventType.BOUNCED, event_data.get("severity")), event_type
        )

        # Use signature.token for event_id, rather than event_data.id,
        # because the latter is only "guaranteed to be unique within a day".
        event_id = esp_event.get("signature", {}).get("token")

        recipient = event_data.get("recipient")

        try:
            timestamp = datetime.fromtimestamp(
                float(event_data["timestamp"]), tz=timezone.utc
            )
        except KeyError:
            timestamp = None

        try:
            message_id = event_data["message"]["headers"]["message-id"]
        except KeyError:
            message_id = None
        if message_id and not message_id.startswith("<"):
            message_id = "<{}>".format(message_id)

        metadata = event_data.get("user-variables", {})
        tags = event_data.get("tags", [])

        try:
            delivery_status = event_data["delivery-status"]
        except KeyError:
            description = None
            mta_response = None
        else:
            description = delivery_status.get("description")
            mta_response = delivery_status.get("message")

        if "reason" in event_data:
            reject_reason = self.reject_reasons.get(
                event_data["reason"], RejectReason.OTHER
            )
        else:
            reject_reason = None

        if event_type == EventType.REJECTED:
            # This event has a somewhat different structure than the others...
            description = description or event_data.get("reject", {}).get("reason")
            reject_reason = reject_reason or RejectReason.OTHER
            if not recipient:
                try:
                    to_email = parse_single_address(
                        event_data["message"]["headers"]["to"]
                    )
                except (AnymailInvalidAddress, KeyError):
                    pass
                else:
                    recipient = to_email.addr_spec

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            message_id=message_id,
            event_id=event_id,
            recipient=recipient,
            reject_reason=reject_reason,
            description=description,
            mta_response=mta_response,
            tags=tags,
            metadata=metadata,
            click_url=event_data.get("url"),
            user_agent=event_data.get("client-info", {}).get("user-agent"),
            esp_event=esp_event,
        )

    # Legacy event handling
    # (Prior to 2018-06-29, these were the only Mailgun events.)

    legacy_event_types = {
        # Map Mailgun event: Anymail normalized type
        "delivered": EventType.DELIVERED,
        "dropped": EventType.REJECTED,
        "bounced": EventType.BOUNCED,
        "complained": EventType.COMPLAINED,
        "unsubscribed": EventType.UNSUBSCRIBED,
        "opened": EventType.OPENED,
        "clicked": EventType.CLICKED,
        # Mailgun does not send events corresponding to QUEUED or DEFERRED
    }

    legacy_reject_reasons = {
        # Map Mailgun (SMTP) error codes to Anymail normalized reject_reason.
        # By default, we will treat anything 400-599 as REJECT_BOUNCED
        # so only exceptions are listed here.
        499: RejectReason.TIMED_OUT,  # unable to connect to MX
        # (499 also covers invalid recipients)
        # These 6xx codes appear to be Mailgun extensions to SMTP
        # (and don't seem to be documented anywhere):
        605: RejectReason.BOUNCED,  # previous bounce
        607: RejectReason.SPAM,  # previous spam complaint
    }

    def mailgun_legacy_to_anymail_event(self, esp_event):
        # esp_event is a Django QueryDict (from request.POST),
        # which has multi-valued fields, but is *not* case-insensitive.
        # Because of the way Mailgun merges user-variables into the event,
        # we must generally use the *first* value of any multi-valued field
        # to avoid potential conflicting user-data.
        esp_event.getfirst = querydict_getfirst.__get__(esp_event)

        if "event" not in esp_event and "sender" in esp_event:
            # Inbound events don't (currently) have an event field
            raise AnymailConfigurationError(
                "You seem to have set Mailgun's *inbound* route "
                "to Anymail's Mailgun *tracking* webhook URL."
            )

        event_type = self.legacy_event_types.get(
            esp_event.getfirst("event"), EventType.UNKNOWN
        )
        # use *last* value of timestamp:
        timestamp = datetime.fromtimestamp(int(esp_event["timestamp"]), tz=timezone.utc)
        # Message-Id is not documented for every event, but seems to always be included.
        # (It's sometimes spelled as 'message-id', lowercase, and missing the
        # <angle-brackets>.)
        message_id = esp_event.getfirst("Message-Id", None) or esp_event.getfirst(
            "message-id", None
        )
        if message_id and not message_id.startswith("<"):
            message_id = "<{}>".format(message_id)

        description = esp_event.getfirst("description", None)
        mta_response = esp_event.getfirst("error", None) or esp_event.getfirst(
            "notification", None
        )
        reject_reason = None
        try:
            mta_status = int(esp_event.getfirst("code"))
        except (KeyError, TypeError):
            pass
        except ValueError:
            # RFC-3463 extended SMTP status code
            # (class.subject.detail, where class is "2", "4" or "5")
            try:
                status_class = esp_event.getfirst("code").split(".")[0]
            except (TypeError, IndexError):
                # illegal SMTP status code format
                pass
            else:
                reject_reason = (
                    RejectReason.BOUNCED
                    if status_class in ("4", "5")
                    else RejectReason.OTHER
                )
        else:
            reject_reason = self.legacy_reject_reasons.get(
                mta_status,
                RejectReason.BOUNCED if 400 <= mta_status < 600 else RejectReason.OTHER,
            )

        metadata = self._extract_legacy_metadata(esp_event)

        # tags are supposed to be in 'tag' fields,
        # but are sometimes in undocumented X-Mailgun-Tag
        tags = esp_event.getlist("tag", None) or esp_event.getlist("X-Mailgun-Tag", [])

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            message_id=message_id,
            event_id=esp_event.get("token", None),  # use *last* value of token
            recipient=esp_event.getfirst("recipient", None),
            reject_reason=reject_reason,
            description=description,
            mta_response=mta_response,
            tags=tags,
            metadata=metadata,
            click_url=esp_event.getfirst("url", None),
            user_agent=esp_event.getfirst("user-agent", None),
            esp_event=esp_event,
        )

    def _extract_legacy_metadata(self, esp_event):
        # Mailgun merges user-variables into the POST fields. If you know which user
        # variable you want to retrieve--and it doesn't conflict with a Mailgun event
        # field--that's fine. But if you want to extract all user-variables (like we
        # do), it's more complicated...
        event_type = esp_event.getfirst("event")
        metadata = {}

        if "message-headers" in esp_event:
            # For events where original message headers are available, it's most
            # reliable to recover user-variables from the X-Mailgun-Variables header(s).
            headers = json.loads(esp_event["message-headers"])
            variables = [
                value for [field, value] in headers if field == "X-Mailgun-Variables"
            ]
            if len(variables) >= 1:
                # Each X-Mailgun-Variables value is JSON. Parse and merge them all into
                # single dict:
                metadata = merge_dicts_shallow(
                    *[json.loads(value) for value in variables]
                )

        elif event_type in self._known_legacy_event_fields:
            # For other events, we must extract from the POST fields, ignoring known
            # Mailgun event parameters, and treating all other values as user-variables.
            known_fields = self._known_legacy_event_fields[event_type]
            for field, values in esp_event.lists():
                if field not in known_fields:
                    # Unknown fields are assumed to be user-variables. (There should
                    # really only be a single value, but just in case take the last one
                    # to match QueryDict semantics.)
                    metadata[field] = values[-1]
                elif field == "tag":
                    # There's no way to distinguish a user-variable named 'tag' from
                    # an actual tag, so don't treat this/these value(s) as metadata.
                    pass
                elif len(values) == 1:
                    # This is an expected event parameter, and since there's only a
                    # single value it must be the event param, not metadata.
                    pass
                else:
                    # This is an expected event parameter, but there are (at least) two
                    # values. One is the event param, and the other is a user-variable
                    # metadata value. Which is which depends on the field:
                    if field in {"signature", "timestamp", "token"}:
                        # values = [user-variable, event-param]
                        metadata[field] = values[0]
                    else:
                        # values = [event-param, user-variable]
                        metadata[field] = values[-1]

        return metadata

    _common_legacy_event_fields = {
        # These fields are documented to appear in all Mailgun
        # opened, clicked and unsubscribed events:
        "event",
        "recipient",
        "domain",
        "ip",
        "country",
        "region",
        "city",
        "user-agent",
        "device-type",
        "client-type",
        "client-name",
        "client-os",
        "campaign-id",
        "campaign-name",
        "tag",
        "mailing-list",
        "timestamp",
        "token",
        "signature",
        # Undocumented, but observed in actual events:
        "body-plain",
        "h",
        "message-id",
    }
    _known_legacy_event_fields = {
        # For all Mailgun event types that *don't* include message-headers,
        # map Mailgun (not normalized) event type to set of expected event fields.
        # Used for metadata extraction.
        "clicked": _common_legacy_event_fields | {"url"},
        "opened": _common_legacy_event_fields,
        "unsubscribed": _common_legacy_event_fields,
    }


class MailgunInboundWebhookView(MailgunBaseWebhookView):
    """Handler for Mailgun inbound (route forward-to-url) webhook"""

    signal = inbound

    def parse_events(self, request):
        if request.content_type == "application/json":
            esp_event = json.loads(request.body.decode("utf-8"))
            event_type = esp_event.get("event-data", {}).get("event", "")
            raise AnymailConfigurationError(
                "You seem to have set Mailgun's *%s tracking* webhook "
                "to Anymail's Mailgun *inbound* webhook URL. "
                "(Or Mailgun has changed inbound events to use json.)" % event_type
            )
        return [self.esp_to_anymail_event(request)]

    def esp_to_anymail_event(self, request):
        # Inbound uses the entire Django request as esp_event, because
        # we need POST and FILES. Note that request.POST is case-sensitive
        # (unlike email.message.Message headers).
        esp_event = request

        if request.POST.get("event", "inbound") != "inbound":
            # (Legacy) tracking event
            raise AnymailConfigurationError(
                "You seem to have set Mailgun's *%s tracking* webhook "
                "to Anymail's Mailgun *inbound* webhook URL." % request.POST["event"]
            )

        if "attachments" in request.POST:
            # Inbound route used store() rather than forward().
            # ("attachments" seems to be the only POST param that differs between
            # store and forward; Anymail could support store by handling the JSON
            # attachments param in message_from_mailgun_parsed.)
            raise AnymailConfigurationError(
                "You seem to have configured Mailgun's receiving route using"
                " the store() action. Anymail's inbound webhook requires"
                " the forward() action."
            )

        if "body-mime" in request.POST:
            # Raw-MIME
            message = AnymailInboundMessage.parse_raw_mime(request.POST["body-mime"])
        else:
            # Fully-parsed
            message = self.message_from_mailgun_parsed(request)

        message.envelope_sender = request.POST.get("sender", None)
        message.envelope_recipient = request.POST.get("recipient", None)
        message.stripped_text = request.POST.get("stripped-text", None)
        message.stripped_html = request.POST.get("stripped-html", None)

        message.spam_detected = message.get("X-Mailgun-Sflag", "No").lower() == "yes"
        try:
            message.spam_score = float(message["X-Mailgun-Sscore"])
        except (TypeError, ValueError):
            pass

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=datetime.fromtimestamp(
                int(request.POST["timestamp"]), tz=timezone.utc
            ),
            event_id=request.POST.get("token", None),
            esp_event=esp_event,
            message=message,
        )

    def message_from_mailgun_parsed(self, request):
        """Construct a Message from Mailgun's "fully-parsed" fields"""
        # Mailgun transcodes all fields to UTF-8 for "fully parsed" messages
        try:
            attachment_count = int(request.POST["attachment-count"])
        except (KeyError, TypeError):
            attachments = None
        else:
            # Load attachments from posted files: attachment-1, attachment-2, etc.
            # content-id-map is {content-id: attachment-id}, identifying which files
            # are inline attachments. Invert it to {attachment-id: content-id}, while
            # handling potentially duplicate content-ids.
            field_to_content_id = json.loads(
                request.POST.get("content-id-map", "{}"),
                object_pairs_hook=lambda pairs: {
                    att_id: cid for (cid, att_id) in pairs
                },
            )
            attachments = []
            for n in range(1, attachment_count + 1):
                attachment_id = "attachment-%d" % n
                try:
                    file = request.FILES[attachment_id]
                except KeyError:
                    # Django's multipart/form-data handling drops FILES with certain
                    # filenames (for security) or with empty filenames (Django ticket
                    # 15879).
                    # (To avoid this problem, use Mailgun's "raw MIME" inbound option.)
                    pass
                else:
                    content_id = field_to_content_id.get(attachment_id)
                    attachment = (
                        AnymailInboundMessage.construct_attachment_from_uploaded_file(
                            file, content_id=content_id
                        )
                    )
                    attachments.append(attachment)

        return AnymailInboundMessage.construct(
            # message-headers includes From, To, Cc, Subject, etc.
            headers=json.loads(request.POST["message-headers"]),
            text=request.POST.get("body-plain", None),
            html=request.POST.get("body-html", None),
            attachments=attachments,
        )



import json
from datetime import datetime, timezone
from email.utils import unquote
from urllib.parse import quote, urljoin

import requests

from ..exceptions import AnymailConfigurationError
from ..inbound import AnymailInboundMessage
from ..signals import (
    AnymailInboundEvent,
    AnymailTrackingEvent,
    EventType,
    RejectReason,
    inbound,
    tracking,
)
from ..utils import get_anymail_setting
from .base import AnymailBaseWebhookView


class SendinBlueBaseWebhookView(AnymailBaseWebhookView):
    esp_name = "SendinBlue"


class SendinBlueTrackingWebhookView(SendinBlueBaseWebhookView):
    """Handler for SendinBlue delivery and engagement tracking webhooks"""

    signal = tracking

    def parse_events(self, request):
        esp_event = json.loads(request.body.decode("utf-8"))
        if "items" in esp_event:
            # This is an inbound webhook post
            raise AnymailConfigurationError(
                "You seem to have set SendinBlue's *inbound* webhook URL "
                "to Anymail's SendinBlue *tracking* webhook URL."
            )
        return [self.esp_to_anymail_event(esp_event)]

    # SendinBlue's webhook payload data doesn't seem to be documented anywhere.
    # There's a list of webhook events at https://apidocs.sendinblue.com/webhooks/#3.
    event_types = {
        # Map SendinBlue event type: Anymail normalized (event type, reject reason)
        # received even if message won't be sent (e.g., before "blocked"):
        "request": (EventType.QUEUED, None),
        "delivered": (EventType.DELIVERED, None),
        "hard_bounce": (EventType.BOUNCED, RejectReason.BOUNCED),
        "soft_bounce": (EventType.BOUNCED, RejectReason.BOUNCED),
        "blocked": (EventType.REJECTED, RejectReason.BLOCKED),
        "spam": (EventType.COMPLAINED, RejectReason.SPAM),
        "invalid_email": (EventType.BOUNCED, RejectReason.INVALID),
        "deferred": (EventType.DEFERRED, None),
        "opened": (EventType.OPENED, None),  # see also unique_opened below
        "click": (EventType.CLICKED, None),
        "unsubscribe": (EventType.UNSUBSCRIBED, None),
        # shouldn't occur for transactional messages:
        "list_addition": (EventType.SUBSCRIBED, None),
        "unique_opened": (EventType.OPENED, None),  # you'll *also* receive an "opened"
    }

    def esp_to_anymail_event(self, esp_event):
        esp_type = esp_event.get("event")
        event_type, reject_reason = self.event_types.get(
            esp_type, (EventType.UNKNOWN, None)
        )
        recipient = esp_event.get("email")

        try:
            # SendinBlue supplies "ts", "ts_event" and "date" fields, which seem to be
            # based on the timezone set in the account preferences (and possibly with
            # inconsistent DST adjustment). "ts_epoch" is the only field that seems to
            # be consistently UTC; it's in milliseconds
            timestamp = datetime.fromtimestamp(
                esp_event["ts_epoch"] / 1000.0, tz=timezone.utc
            )
        except (KeyError, ValueError):
            timestamp = None

        tags = []
        try:
            # If `tags` param set on send, webhook payload includes 'tags' array field.
            tags = esp_event["tags"]
        except KeyError:
            try:
                # If `X-Mailin-Tag` header set on send, webhook payload includes single
                # 'tag' string. (If header not set, webhook 'tag' will be the template
                # name for template sends.)
                tags = [esp_event["tag"]]
            except KeyError:
                pass

        try:
            metadata = json.loads(esp_event["X-Mailin-custom"])
        except (KeyError, TypeError):
            metadata = {}

        return AnymailTrackingEvent(
            description=None,
            esp_event=esp_event,
            # SendinBlue doesn't provide a unique event id:
            event_id=None,
            event_type=event_type,
            message_id=esp_event.get("message-id"),
            metadata=metadata,
            mta_response=esp_event.get("reason"),
            recipient=recipient,
            reject_reason=reject_reason,
            tags=tags,
            timestamp=timestamp,
            user_agent=None,
            click_url=esp_event.get("link"),
        )


class SendinBlueInboundWebhookView(SendinBlueBaseWebhookView):
    """Handler for SendinBlue inbound email webhooks"""

    signal = inbound

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # API is required to fetch inbound attachment content:
        self.api_key = get_anymail_setting(
            "api_key",
            esp_name=self.esp_name,
            kwargs=kwargs,
            allow_bare=True,
        )
        self.api_url = get_anymail_setting(
            "api_url",
            esp_name=self.esp_name,
            kwargs=kwargs,
            default="https://api.brevo.com/v3/",
        )
        if not self.api_url.endswith("/"):
            self.api_url += "/"

    def parse_events(self, request):
        payload = json.loads(request.body.decode("utf-8"))
        try:
            esp_events = payload["items"]
        except KeyError:
            # This is not n inbound webhook post
            raise AnymailConfigurationError(
                "You seem to have set SendinBlue's *tracking* webhook URL "
                "to Anymail's SendinBlue *inbound* webhook URL."
            )
        else:
            return [self.esp_to_anymail_event(esp_event) for esp_event in esp_events]

    def esp_to_anymail_event(self, esp_event):
        # Inbound event's "Uuid" is documented as
        # "A list of recipients UUID (can be used with the Public API)".
        # In practice, it seems to be a single-item list (even when sending
        # to multiple inbound recipients at once) that uniquely identifies this
        # inbound event. (And works as a param for the /inbound/events/{uuid} API
        # that will "Fetch all events history for one particular received email.")
        try:
            event_id = esp_event["Uuid"][0]
        except (KeyError, IndexError):
            event_id = None

        attachments = [
            self._fetch_attachment(attachment)
            for attachment in esp_event.get("Attachments", [])
        ]
        headers = [
            (name, value)
            for name, values in esp_event.get("Headers", {}).items()
            # values is string if single header instance, list of string if multiple
            for value in ([values] if isinstance(values, str) else values)
        ]

        # (esp_event From, To, Cc, ReplyTo, Subject, Date, etc. are also in Headers)
        message = AnymailInboundMessage.construct(
            headers=headers,
            text=esp_event.get("RawTextBody", ""),
            html=esp_event.get("RawHtmlBody", ""),
            attachments=attachments,
        )

        if message["Return-Path"]:
            message.envelope_sender = unquote(message["Return-Path"])
        if message["Delivered-To"]:
            message.envelope_recipient = unquote(message["Delivered-To"])
        message.stripped_text = esp_event.get("ExtractedMarkdownMessage")

        # Documented as "Spam.Score" object, but both example payload
        # and actual received payload use single "SpamScore" field:
        message.spam_score = esp_event.get("SpamScore")

        return AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=None,  # Brevo doesn't provide inbound event timestamp
            event_id=event_id,
            esp_event=esp_event,
            message=message,
        )

    def _fetch_attachment(self, attachment):
        # Download attachment content from SendinBlue API.
        # FUTURE: somehow defer download until attachment is accessed?
        token = attachment["DownloadToken"]
        url = urljoin(self.api_url, f"inbound/attachments/{quote(token, safe='')}")
        response = requests.get(url, headers={"api-key": self.api_key})
        response.raise_for_status()  # or maybe just log and continue?

        content = response.content
        # Prefer response Content-Type header to attachment ContentType field,
        # as the header will include charset but the ContentType field won't.
        content_type = response.headers.get("Content-Type") or attachment["ContentType"]
        return AnymailInboundMessage.construct_attachment(
            content_type=content_type,
            content=content,
            filename=attachment.get("Name"),
            content_id=attachment.get("ContentID"),
        )



import warnings
from email.charset import QP, Charset
from email.mime.text import MIMEText

from ..exceptions import (
    AnymailAPIError,
    AnymailDeprecationWarning,
    AnymailImproperlyInstalled,
)
from ..message import AnymailRecipientStatus
from ..utils import UNSET, get_anymail_setting
from .base import AnymailBaseBackend, BasePayload

try:
    import boto3
    from botocore.exceptions import BotoCoreError, ClientError, ConnectionError
except ImportError as err:
    raise AnymailImproperlyInstalled(
        missing_package="boto3", install_extra="amazon-ses"
    ) from err


# boto3 has several root exception classes; this is meant to cover all of them
BOTO_BASE_ERRORS = (BotoCoreError, ClientError, ConnectionError)


class EmailBackend(AnymailBaseBackend):
    """
    Amazon SES v1 Email Backend (using boto3)
    """

    esp_name = "Amazon SES"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        from .amazon_ses import _get_anymail_boto3_params

        warnings.warn(
            "anymail.backends.amazon_sesv1.EmailBackend is deprecated"
            " and will be removed in the near future. Please migrate"
            " to anymail.backends.amazon_ses.EmailBackend using Amazon SES v2.",
            AnymailDeprecationWarning,
        )
        super().__init__(**kwargs)

        # AMAZON_SES_CLIENT_PARAMS is optional
        # (boto3 can find credentials several other ways)
        self.session_params, self.client_params = _get_anymail_boto3_params(
            kwargs=kwargs
        )
        self.configuration_set_name = get_anymail_setting(
            "configuration_set_name",
            esp_name=self.esp_name,
            kwargs=kwargs,
            allow_bare=False,
            default=None,
        )
        self.message_tag_name = get_anymail_setting(
            "message_tag_name",
            esp_name=self.esp_name,
            kwargs=kwargs,
            allow_bare=False,
            default=None,
        )
        self.client = None

    def open(self):
        if self.client:
            return False  # already exists
        try:
            self.client = boto3.session.Session(**self.session_params).client(
                "ses", **self.client_params
            )
        except Exception:
            if not self.fail_silently:
                raise
        else:
            return True  # created client

    def close(self):
        if self.client is None:
            return
        # self.client.close()  # boto3 doesn't support (or require) client shutdown
        self.client = None

    def _send(self, message):
        if self.client:
            return super()._send(message)
        elif self.fail_silently:
            # (Probably missing boto3 credentials in open().)
            return False
        else:
            class_name = self.__class__.__name__
            raise RuntimeError(
                "boto3 Session has not been opened in {class_name}._send. "
                "(This is either an implementation error in {class_name}, "
                "or you are incorrectly calling _send directly.)".format(
                    class_name=class_name
                )
            )

    def build_message_payload(self, message, defaults):
        # The SES SendRawEmail and SendBulkTemplatedEmail calls have
        # very different signatures, so use a custom payload for each
        if getattr(message, "template_id", UNSET) is not UNSET:
            return AmazonSESSendBulkTemplatedEmailPayload(message, defaults, self)
        else:
            return AmazonSESSendRawEmailPayload(message, defaults, self)

    def post_to_esp(self, payload, message):
        try:
            response = payload.call_send_api(self.client)
        except BOTO_BASE_ERRORS as err:
            # ClientError has a response attr with parsed json error response
            # (other errors don't)
            raise AnymailAPIError(
                str(err),
                backend=self,
                email_message=message,
                payload=payload,
                response=getattr(err, "response", None),
            ) from err
        return response

    def parse_recipient_status(self, response, payload, message):
        return payload.parse_recipient_status(response)


class AmazonSESBasePayload(BasePayload):
    def init_payload(self):
        self.params = {}
        if self.backend.configuration_set_name is not None:
            self.params["ConfigurationSetName"] = self.backend.configuration_set_name

    def call_send_api(self, ses_client):
        raise NotImplementedError()

    def parse_recipient_status(self, response):
        # response is the parsed (dict) JSON returned from the API call
        raise NotImplementedError()

    def set_esp_extra(self, extra):
        # e.g., ConfigurationSetName, FromArn, SourceArn, ReturnPathArn
        self.params.update(extra)


class AmazonSESSendRawEmailPayload(AmazonSESBasePayload):
    def init_payload(self):
        super().init_payload()
        self.all_recipients = []
        self.mime_message = self.message.message()

        # Work around an Amazon SES bug where, if all of:
        #   - the message body (text or html) contains non-ASCII characters
        #   - the body is sent with `Content-Transfer-Encoding: 8bit`
        #     (which is Django email's default for most non-ASCII bodies)
        #   - you are using an SES ConfigurationSet with open or click tracking enabled
        # then SES replaces the non-ASCII characters with question marks as it rewrites
        # the message to add tracking. Forcing `CTE: quoted-printable` avoids the
        # problem. (https://forums.aws.amazon.com/thread.jspa?threadID=287048)
        for part in self.mime_message.walk():
            if (
                part.get_content_maintype() == "text"
                and part["Content-Transfer-Encoding"] == "8bit"
            ):
                content = part.get_payload()
                del part["Content-Transfer-Encoding"]
                qp_charset = Charset(part.get_content_charset("us-ascii"))
                qp_charset.body_encoding = QP
                # (can't use part.set_payload, because SafeMIMEText can undo
                # this workaround)
                MIMEText.set_payload(part, content, charset=qp_charset)

    def call_send_api(self, ses_client):
        # Set Destinations to make sure we pick up all recipients (including bcc).
        # Any non-ASCII characters in recipient domains must be encoded with Punycode.
        # (Amazon SES doesn't support non-ASCII recipient usernames.)
        self.params["Destinations"] = [email.address for email in self.all_recipients]
        self.params["RawMessage"] = {"Data": self.mime_message.as_bytes()}
        return ses_client.send_raw_email(**self.params)

    def parse_recipient_status(self, response):
        try:
            message_id = response["MessageId"]
        except (KeyError, TypeError) as err:
            raise AnymailAPIError(
                "%s parsing Amazon SES send result %r" % (str(err), response),
                backend=self.backend,
                email_message=self.message,
                payload=self,
            ) from None

        recipient_status = AnymailRecipientStatus(
            message_id=message_id, status="queued"
        )
        return {
            recipient.addr_spec: recipient_status for recipient in self.all_recipients
        }

    # Standard EmailMessage attrs...
    # These all get rolled into the RFC-5322 raw mime directly via
    # EmailMessage.message()

    def _no_send_defaults(self, attr):
        # Anymail global send defaults don't work for standard attrs, because the
        # merged/computed value isn't forced back into the EmailMessage.
        if attr in self.defaults:
            self.unsupported_feature(
                "Anymail send defaults for '%s' with Amazon SES" % attr
            )

    def set_from_email_list(self, emails):
        # Although Amazon SES will send messages with any From header, it can only parse
        # Source if the From header is a single email. Explicit Source avoids an
        # "Illegal address" error:
        if len(emails) > 1:
            self.params["Source"] = emails[0].addr_spec
        # (else SES will look at the (single) address in the From header)

    def set_recipients(self, recipient_type, emails):
        self.all_recipients += emails
        # included in mime_message
        assert recipient_type in ("to", "cc", "bcc")
        self._no_send_defaults(recipient_type)

    def set_subject(self, subject):
        # included in mime_message
        self._no_send_defaults("subject")

    def set_reply_to(self, emails):
        # included in mime_message
        self._no_send_defaults("reply_to")

    def set_extra_headers(self, headers):
        # included in mime_message
        self._no_send_defaults("extra_headers")

    def set_text_body(self, body):
        # included in mime_message
        self._no_send_defaults("body")

    def set_html_body(self, body):
        # included in mime_message
        self._no_send_defaults("body")

    def set_alternatives(self, alternatives):
        # included in mime_message
        self._no_send_defaults("alternatives")

    def set_attachments(self, attachments):
        # included in mime_message
        self._no_send_defaults("attachments")

    # Anymail-specific payload construction
    def set_envelope_sender(self, email):
        self.params["Source"] = email.addr_spec

    def set_spoofed_to_header(self, header_to):
        # django.core.mail.EmailMessage.message() has already set
        #   self.mime_message["To"] = header_to
        # and performed any necessary header sanitization.
        #
        # The actual "to" is already in self.all_recipients,
        # which is used as the SendRawEmail Destinations later.
        #
        # So, nothing to do here, except prevent the default
        # "unsupported feature" error.
        pass

    def set_metadata(self, metadata):
        # Amazon SES has two mechanisms for adding custom data to a message:
        # * Custom message headers are available to webhooks (SNS notifications),
        #   but not in CloudWatch metrics/dashboards or Kinesis Firehose streams.
        #   Custom headers can be sent only with SendRawEmail.
        # * "Message Tags" are available to CloudWatch and Firehose, and to SNS
        #   notifications for SES *events* but not SES *notifications*. (Got that?)
        #   Message Tags also allow *very* limited characters in both name and value.
        #   Message Tags can be sent with any SES send call.
        # (See "How do message tags work?" in
        # https://aws.amazon.com/blogs/ses/introducing-sending-metrics/
        # and https://forums.aws.amazon.com/thread.jspa?messageID=782922.)
        # To support reliable retrieval in webhooks, just use custom headers for
        # metadata.
        self.mime_message["X-Metadata"] = self.serialize_json(metadata)

    def set_tags(self, tags):
        # See note about Amazon SES Message Tags and custom headers in set_metadata
        # above. To support reliable retrieval in webhooks, use custom headers for tags.
        # (There are no restrictions on number or content for custom header tags.)
        for tag in tags:
            # creates multiple X-Tag headers, one per tag:
            self.mime_message.add_header("X-Tag", tag)

        # Also *optionally* pass a single Message Tag if the AMAZON_SES_MESSAGE_TAG_NAME
        # Anymail setting is set (default no). The AWS API restricts tag content in this
        # case. (This is useful for dashboard segmentation; use esp_extra["Tags"] for
        # anything more complex.)
        if tags and self.backend.message_tag_name is not None:
            if len(tags) > 1:
                self.unsupported_feature(
                    "multiple tags with the AMAZON_SES_MESSAGE_TAG_NAME setting"
                )
            self.params.setdefault("Tags", []).append(
                {"Name": self.backend.message_tag_name, "Value": tags[0]}
            )

    def set_template_id(self, template_id):
        raise NotImplementedError(
            "AmazonSESSendRawEmailPayload should not have been used with template_id"
        )

    def set_merge_data(self, merge_data):
        self.unsupported_feature("merge_data without template_id")

    def set_merge_global_data(self, merge_global_data):
        self.unsupported_feature("global_merge_data without template_id")


class AmazonSESSendBulkTemplatedEmailPayload(AmazonSESBasePayload):
    def init_payload(self):
        super().init_payload()
        # late-bind recipients and merge_data in call_send_api
        self.recipients = {"to": [], "cc": [], "bcc": []}
        self.merge_data = {}

    def call_send_api(self, ses_client):
        # include any 'cc' or 'bcc' in every destination
        cc_and_bcc_addresses = {}
        if self.recipients["cc"]:
            cc_and_bcc_addresses["CcAddresses"] = [
                cc.address for cc in self.recipients["cc"]
            ]
        if self.recipients["bcc"]:
            cc_and_bcc_addresses["BccAddresses"] = [
                bcc.address for bcc in self.recipients["bcc"]
            ]

        # set up destination and data for each 'to'
        self.params["Destinations"] = [
            {
                "Destination": dict(ToAddresses=[to.address], **cc_and_bcc_addresses),
                "ReplacementTemplateData": self.serialize_json(
                    self.merge_data.get(to.addr_spec, {})
                ),
            }
            for to in self.recipients["to"]
        ]

        return ses_client.send_bulk_templated_email(**self.params)

    def parse_recipient_status(self, response):
        try:
            # response["Status"] should be a list in Destinations (to) order
            anymail_statuses = [
                AnymailRecipientStatus(
                    message_id=status.get("MessageId", None),
                    status="queued" if status.get("Status") == "Success" else "failed",
                )
                for status in response["Status"]
            ]
        except (KeyError, TypeError) as err:
            raise AnymailAPIError(
                "%s parsing Amazon SES send result %r" % (str(err), response),
                backend=self.backend,
                email_message=self.message,
                payload=self,
            ) from None

        to_addrs = [to.addr_spec for to in self.recipients["to"]]
        if len(anymail_statuses) != len(to_addrs):
            raise AnymailAPIError(
                "Sent to %d destinations, but only %d statuses in Amazon SES"
                " send result %r" % (len(to_addrs), len(anymail_statuses), response),
                backend=self.backend,
                email_message=self.message,
                payload=self,
            )

        return dict(zip(to_addrs, anymail_statuses))

    def set_from_email(self, email):
        # this will RFC2047-encode display_name if needed:
        self.params["Source"] = email.address

    def set_recipients(self, recipient_type, emails):
        # late-bound in call_send_api
        assert recipient_type in ("to", "cc", "bcc")
        self.recipients[recipient_type] = emails

    def set_subject(self, subject):
        # (subject can only come from template; you can use substitution vars in that)
        if subject:
            self.unsupported_feature("overriding template subject")

    def set_reply_to(self, emails):
        if emails:
            self.params["ReplyToAddresses"] = [email.address for email in emails]

    def set_extra_headers(self, headers):
        self.unsupported_feature("extra_headers with template")

    def set_text_body(self, body):
        if body:
            self.unsupported_feature("overriding template body content")

    def set_html_body(self, body):
        if body:
            self.unsupported_feature("overriding template body content")

    def set_attachments(self, attachments):
        if attachments:
            self.unsupported_feature("attachments with template")

    # Anymail-specific payload construction
    def set_envelope_sender(self, email):
        self.params["ReturnPath"] = email.addr_spec

    def set_metadata(self, metadata):
        # no custom headers with SendBulkTemplatedEmail
        self.unsupported_feature("metadata with template")

    def set_tags(self, tags):
        # no custom headers with SendBulkTemplatedEmail, but support
        # AMAZON_SES_MESSAGE_TAG_NAME if used (see tags/metadata in
        # AmazonSESSendRawEmailPayload for more info)
        if tags:
            if self.backend.message_tag_name is not None:
                if len(tags) > 1:
                    self.unsupported_feature(
                        "multiple tags with the AMAZON_SES_MESSAGE_TAG_NAME setting"
                    )
                self.params["DefaultTags"] = [
                    {"Name": self.backend.message_tag_name, "Value": tags[0]}
                ]
            else:
                self.unsupported_feature(
                    "tags with template (unless using the"
                    " AMAZON_SES_MESSAGE_TAG_NAME setting)"
                )

    def set_template_id(self, template_id):
        self.params["Template"] = template_id

    def set_merge_data(self, merge_data):
        # late-bound in call_send_api
        self.merge_data = merge_data

    def set_merge_global_data(self, merge_global_data):
        self.params["DefaultTemplateData"] = self.serialize_json(merge_global_data)



import warnings
from base64 import b64decode
from email.message import EmailMessage
from email.parser import BytesParser, Parser
from email.policy import default as default_policy
from email.utils import unquote

from django.core.files.uploadedfile import SimpleUploadedFile

from .exceptions import AnymailDeprecationWarning
from .utils import angle_wrap, parse_address_list, parse_rfc2822date


class AnymailInboundMessage(EmailMessage):
    """
    A normalized, parsed inbound email message.

    A subclass of email.message.EmailMessage, with some additional
    convenience properties.
    """

    # Why Python email.message.EmailMessage rather than django.core.mail.EmailMessage?
    # Django's EmailMessage is really intended for constructing a (limited subset of)
    # an EmailMessage to send; Python's EmailMessage is better designed for representing
    # arbitrary messages:
    #
    # * Python's EmailMessage is easily parsed from raw mime (which is an inbound format
    #   provided by many ESPs), and can accurately represent any mime email received
    # * Python's EmailMessage can represent repeated header fields (e.g., "Received")
    #   which are common in inbound messages
    # * Django's EmailMessage defaults a bunch of properties in ways that aren't helpful
    #   (e.g., from_email from settings)

    def __init__(self, *args, **kwargs):
        # Note: this must accept zero arguments,
        # for use with message_from_string (email.parser)
        super().__init__(*args, **kwargs)

        # Additional attrs provided by some ESPs:
        self.envelope_sender = None
        self.envelope_recipient = None
        self.stripped_text = None
        self.stripped_html = None
        self.spam_detected = None
        self.spam_score = None

    #
    # Convenience accessors
    #

    @property
    def from_email(self):
        """EmailAddress"""
        # equivalent to Python 3.2+ message['From'].addresses[0]
        from_email = self.get_address_header("From")
        if len(from_email) == 1:
            return from_email[0]
        elif len(from_email) == 0:
            return None
        else:
            # unusual, but technically-legal multiple-From; preserve list:
            return from_email

    @property
    def to(self):
        """list of EmailAddress objects from To header"""
        # equivalent to Python 3.2+ message['To'].addresses
        return self.get_address_header("To")

    @property
    def cc(self):
        """list of EmailAddress objects from Cc header"""
        # equivalent to Python 3.2+ message['Cc'].addresses
        return self.get_address_header("Cc")

    @property
    def bcc(self):
        """list of EmailAddress objects from Bcc header"""
        # equivalent to Python 3.2+ message['Bcc'].addresses
        return self.get_address_header("Bcc")

    @property
    def subject(self):
        """str value of Subject header, or None"""
        return self["Subject"]

    @property
    def date(self):
        """datetime.datetime from Date header, or None if missing/invalid"""
        # equivalent to Python 3.2+ message['Date'].datetime
        return self.get_date_header("Date")

    @property
    def text(self):
        """Contents of the (first) text/plain body part, or None"""
        return self._get_body_content("text/plain")

    @property
    def html(self):
        """Contents of the (first) text/html body part, or None"""
        return self._get_body_content("text/html")

    @property
    def attachments(self):
        """list of attachments (as MIMEPart objects); excludes inlines"""
        return [part for part in self.walk() if part.is_attachment()]

    @property
    def inlines(self):
        """list of inline parts (as MIMEPart objects)"""
        return [part for part in self.walk() if part.is_inline()]

    @property
    def inline_attachments(self):
        """DEPRECATED: use content_id_map instead"""
        warnings.warn(
            "inline_attachments has been renamed to content_id_map and will be removed"
            " in the near future.",
            AnymailDeprecationWarning,
        )

        return self.content_id_map

    @property
    def content_id_map(self):
        """dict of Content-ID: attachment (as MIMEPart objects)"""

        return {
            unquote(part["Content-ID"]): part
            for part in self.walk()
            if part.is_inline() and part["Content-ID"] is not None
        }

    def get_address_header(self, header):
        """
        Return the value of header parsed into a (possibly-empty)
        list of EmailAddress objects
        """
        values = self.get_all(header)
        if values is not None:
            if "".join(values).strip() == "":
                values = None
            else:
                values = parse_address_list(values)
        return values or []

    def get_date_header(self, header):
        """Return the value of header parsed into a datetime.date, or None"""
        value = self[header]
        if value is not None:
            value = parse_rfc2822date(value)
        return value

    def _get_body_content(self, content_type):
        # This doesn't handle as many corner cases as Python 3.6
        # email.message.EmailMessage.get_body, but should work correctly
        # for nearly all real-world inbound messages.
        # We're guaranteed to have `is_attachment` available, because all
        # AnymailInboundMessage parts should themselves be AnymailInboundMessage.
        for part in self.walk():
            if part.get_content_type() == content_type and not part.is_attachment():
                return part.get_content_text()
        return None

    def is_inline(self):
        return self.get_content_disposition() == "inline"

    # New for Anymail
    def is_inline_attachment(self):
        """DEPRECATED: use in_inline instead"""
        warnings.warn(
            "is_inline_attachment has been renamed to is_inline and will be removed"
            " in the near future.",
            AnymailDeprecationWarning,
        )

        return self.is_inline()

    def get_content_bytes(self):
        """Return the raw payload bytes"""
        maintype = self.get_content_maintype()
        if maintype == "message":
            # The attachment's payload is a single (parsed) email Message;
            # flatten it to bytes.
            # (Note that self.is_multipart() misleadingly returns True in this case.)
            payload = self.get_payload()
            assert len(payload) == 1  # should be exactly one message
            return payload[0].as_bytes()
        elif maintype == "multipart":
            # The attachment itself is multipart; the payload is a list of parts,
            # and it's not clear which one is the "content".
            raise ValueError(
                "get_content_bytes() is not valid on multipart messages "
                "(perhaps you want as_bytes()?)"
            )
        return self.get_payload(decode=True)

    def get_content_text(self, charset=None, errors=None):
        """Return the payload decoded to text"""
        maintype = self.get_content_maintype()
        if maintype == "message":
            # The attachment's payload is a single (parsed) email Message;
            # flatten it to text.
            # (Note that self.is_multipart() misleadingly returns True in this case.)
            payload = self.get_payload()
            assert len(payload) == 1  # should be exactly one message
            return payload[0].as_string()
        elif maintype == "multipart":
            # The attachment itself is multipart; the payload is a list of parts,
            # and it's not clear which one is the "content".
            raise ValueError(
                "get_content_text() is not valid on multipart messages "
                "(perhaps you want as_string()?)"
            )
        else:
            payload = self.get_payload(decode=True)
            if payload is None:
                return payload
            charset = charset or self.get_content_charset("US-ASCII")
            errors = errors or "replace"
            return payload.decode(charset, errors=errors)

    def as_uploaded_file(self):
        """Return the attachment converted to a Django UploadedFile"""
        if self["Content-Disposition"] is None:
            return None  # this part is not an attachment
        name = self.get_filename()
        content_type = self.get_content_type()
        content = self.get_content_bytes()
        return SimpleUploadedFile(name, content, content_type)

    #
    # Construction
    #
    # These methods are intended primarily for internal Anymail use
    # (in inbound webhook handlers)

    @classmethod
    def parse_raw_mime(cls, s):
        """Returns a new AnymailInboundMessage parsed from str s"""
        if isinstance(s, str):
            # Avoid Python 3.x issue https://bugs.python.org/issue18271
            # (See test_inbound: test_parse_raw_mime_8bit_utf8)
            return cls.parse_raw_mime_bytes(s.encode("utf-8"))
        return Parser(cls, policy=default_policy).parsestr(s)

    @classmethod
    def parse_raw_mime_bytes(cls, b):
        """Returns a new AnymailInboundMessage parsed from bytes b"""
        return BytesParser(cls, policy=default_policy).parsebytes(b)

    @classmethod
    def parse_raw_mime_file(cls, fp):
        """Returns a new AnymailInboundMessage parsed from file-like object fp"""
        if isinstance(fp.read(0), bytes):
            return BytesParser(cls, policy=default_policy).parse(fp)
        else:
            return Parser(cls, policy=default_policy).parse(fp)

    @classmethod
    def construct(
        cls,
        raw_headers=None,
        from_email=None,
        to=None,
        cc=None,
        bcc=None,
        subject=None,
        headers=None,
        text=None,
        text_charset="utf-8",
        html=None,
        html_charset="utf-8",
        attachments=None,
    ):
        """
        Returns a new AnymailInboundMessage constructed from params.

        This is designed to handle the sorts of email fields typically present in ESP
        parsed inbound messages. (It's not a generalized MIME message constructor.)

        :param raw_headers: {str|None}
            base (or complete) message headers as a single string
        :param from_email: {str|None} value for From header
        :param to: {str|None} value for To header
        :param cc: {str|None} value for Cc header
        :param bcc: {str|None} value for Bcc header
        :param subject: {str|None} value for Subject header
        :param headers: {sequence[(str, str)]|mapping|None} additional headers
        :param text: {str|None} plaintext body
        :param text_charset: {str} charset of plaintext body; default utf-8
        :param html: {str|None} html body
        :param html_charset: {str} charset of html body; default utf-8
        :param attachments: {list[MIMEBase]|None} as returned by construct_attachment
        :return: {AnymailInboundMessage}
        """
        if raw_headers is not None:
            msg = Parser(cls, policy=default_policy).parsestr(
                raw_headers, headersonly=True
            )
            # headersonly forces an empty string payload, which breaks things later:
            msg.set_payload(None)
        else:
            msg = cls()

        if from_email is not None:
            del msg["From"]  # override raw_headers value, if any
            msg["From"] = from_email
        if to is not None:
            del msg["To"]
            msg["To"] = to
        if cc is not None:
            del msg["Cc"]
            msg["Cc"] = cc
        if bcc is not None:
            del msg["Bcc"]
            msg["Bcc"] = bcc
        if subject is not None:
            del msg["Subject"]
            msg["Subject"] = subject
        if headers is not None:
            try:
                header_items = headers.items()  # mapping
            except AttributeError:
                header_items = headers  # sequence of (key, value)
            for name, value in header_items:
                msg.add_header(name, value)

        # For simplicity, always build a MIME structure that could support
        # plaintext/html alternative bodies, inline attachments for the body(ies), and
        # message attachments. This may be overkill for simpler messages, but the
        # structure is never incorrect.
        del msg["MIME-Version"]  # override raw_headers values, if any
        del msg["Content-Type"]
        msg["MIME-Version"] = "1.0"
        msg["Content-Type"] = "multipart/mixed"

        related = cls()  # container for alternative bodies and inline attachments
        related["Content-Type"] = "multipart/related"
        msg.attach(related)

        alternatives = cls()  # container for text and html bodies
        alternatives["Content-Type"] = "multipart/alternative"
        related.attach(alternatives)

        if text is not None:
            part = cls()
            part["Content-Type"] = "text/plain"
            part.set_payload(text, charset=text_charset)
            alternatives.attach(part)
        if html is not None:
            part = cls()
            part["Content-Type"] = "text/html"
            part.set_payload(html, charset=html_charset)
            alternatives.attach(part)

        if attachments is not None:
            for attachment in attachments:
                if attachment.is_inline():
                    related.attach(attachment)
                else:
                    msg.attach(attachment)

        return msg

    @classmethod
    def construct_attachment_from_uploaded_file(cls, file, content_id=None):
        # This pulls the entire file into memory; it would be better to implement
        # some sort of lazy attachment where the content is only pulled in if/when
        # requested (and then use file.chunks() to minimize memory usage)
        return cls.construct_attachment(
            content_type=getattr(file, "content_type", None),
            content=file.read(),
            filename=getattr(file, "name", None),
            content_id=content_id,
            charset=getattr(file, "charset", None),
        )

    @classmethod
    def construct_attachment(
        cls,
        content_type,
        content,
        charset=None,
        filename=None,
        content_id=None,
        base64=False,
    ):
        part = cls()
        part["Content-Type"] = content_type
        part["Content-Disposition"] = (
            "inline" if content_id is not None else "attachment"
        )

        if filename is not None:
            part.set_param("name", filename, header="Content-Type")
            part.set_param("filename", filename, header="Content-Disposition")

        if content_id is not None:
            part["Content-ID"] = angle_wrap(content_id)

        if base64:
            content = b64decode(content)

        payload = content
        if part.get_content_maintype() == "message":
            # email.Message parses message/rfc822 parts as a "multipart" (list) payload
            # whose single item is the recursively-parsed message attachment
            if isinstance(content, bytes):
                content = content.decode()
            payload = [cls.parse_raw_mime(content)]
            charset = None

        part.set_payload(payload, charset)
        return part
